#include<linux/module.h>
#include<linux/kthread.h>
#include<linux/wait.h>
#include<linux/mutex.h>

enum thread_index {WAKING_THREAD, FIRST_THREAD, SECOND_THREAD};

static struct thread_structure 
{
	struct task_struct *thread[3];
} threads;

static wait_queue_head_t wait_queue;
static bool condition;
static DEFINE_MUTEX(number_lock);
static int number;

static int first_thread(void *data) 
{
	int counter = 0;
	DEFINE_WAIT(wait);
	for(;;) {
		mutex_lock(&number_lock);
		pr_info("[first_thread] Number value: %d\n",number);
		number++;
		mutex_unlock(&number_lock);
		if(counter%3) {
			add_wait_queue(&wait_queue,&wait);
			while(!condition) {
				prepare_to_wait(&wait_queue,&wait,TASK_INTERRUPTIBLE); 
				if(kthread_should_stop()) 
					return 0;
				printk(KERN_INFO "[first_thread]: awake\n");
				schedule();
			}
			condition=false;
			finish_wait(&wait_queue,&wait); 
		}
		counter++;
	}
}

static int second_thread(void *data) 
{
	int counter = 0;
	DEFINE_WAIT(wait);
	for(;;) {
		mutex_lock(&number_lock);
		pr_info("[second_thread] Number value: %d\n",number);
		number--;
		mutex_unlock(&number_lock);
		if(counter%7) {
			add_wait_queue(&wait_queue,&wait);
				while(!condition) {
					prepare_to_wait(&wait_queue,&wait,TASK_INTERRUPTIBLE); 
					if(kthread_should_stop()) 
						return 0;
					printk(KERN_INFO "[second_thread]: awake\n");
					schedule();
				}
				condition=false;
				finish_wait(&wait_queue,&wait); 
		}
		counter++;
	}
}

static int waking_thread(void *data)
{
	for(;;) {
		if(kthread_should_stop())
			return 0;
		set_current_state(TASK_INTERRUPTIBLE);
		if(schedule_timeout(1*HZ))
			printk(KERN_INFO "Signal received!\n");
		condition=true;
		wake_up_all(&wait_queue);
	}

}

static int __init threads_init(void)
{
	init_waitqueue_head(&wait_queue);
	threads.thread[FIRST_THREAD] = kthread_run(first_thread,NULL,"first_thread");
	threads.thread[SECOND_THREAD] = kthread_run(second_thread,NULL,"second_thread");
	threads.thread[WAKING_THREAD] = kthread_run(waking_thread,NULL,"waking_thread");
	return 0;
}

static void __exit threads_exit(void)
{
	kthread_stop(threads.thread[WAKING_THREAD]);
	kthread_stop(threads.thread[SECOND_THREAD]);
	kthread_stop(threads.thread[FIRST_THREAD]);
}

module_init(threads_init);
module_exit(threads_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("An example of using the Linux kernel threads and a mutex.");
MODULE_AUTHOR("Arkadiusz Chrobot <a.chrobot@tu.kielce.pl>");
